import typing

import ezperm
import django.contrib.auth.mixins


class PermissionRequiredMixin(django.contrib.auth.mixins.PermissionRequiredMixin):
    """Django's view PermissionRequiredMixin with support for ezperm permissions."""

    permission_required: typing.Optional[ezperm.PermissionInterface] = NotImplemented

    def get_permission_required(self) -> typing.Optional[ezperm.PermissionInterface]:
        if self.permission_required is NotImplemented:
            raise ValueError(
                f"`{self.__class__.__name__}` is missing a permission."
                f"Define {self.__class__.__name__}.permission_required, or override"
                f" {self.__class__.__name__}.get_permission_required() method."
            )
        if self.permission_required is None:
            return

        # Handle common errors
        if isinstance(self.permission_required, (tuple, list, set)):
            raise ValueError(
                f"Multiple permissions specified in `{self.__class__.__name__}` view."
                f" Please use & and | operators to combine them into single one."
            )
        if not isinstance(self.permission_required, ezperm.PermissionInterface):
            raise ValueError(
                f"Permission specified in `{self.__class__.__name__}` view is not "
                f"a `{ezperm.PermissionInterface.__name__}` instance."
            )
        return self.permission_required



import typing
import ezperm

if typing.TYPE_CHECKING:
    from django.contrib.auth.models import AbstractUser


# TODO write tests for these


DjangoAdminSitePerms = typing.Union[list[str], str]  # permission types used by the Django admin site


class PermissionsMixin:
    """Django's PermissionsMixin with support for ezperm permissions and admin-compliant permissions."""

    def has_perm(
            self: 'AbstractUser',
            perm: typing.Union[ezperm.PermissionInterface, DjangoAdminSitePerms],
            obj=None,
    ) -> bool:
        """Check if the user has the permission."""
        if not isinstance(perm, ezperm.PermissionInterface):  # handle default Django perms (used in admin)
            return super().has_perm(perm, obj=obj)
        return self.has_perms(perm)

    def has_perms(
            self: 'AbstractUser',
            perms: typing.Optional[typing.Union[ezperm.PermissionInterface, DjangoAdminSitePerms]],
            obj=None,
    ) -> bool:
        """Check if the user has the permission."""
        if not self.is_active:
            return False
        if perms is None:
            return True

        # handle default Django perms (used in admin)
        if not isinstance(perms, ezperm.PermissionInterface):
            if type(perms) is str:
                perms = [perms]
            return super().has_perms(perms, obj=obj)

        if self.is_superuser:
            return True
        return perms(self)



import typing
from .interfaces import PermissionInterface


def permission(func: typing.Callable[[typing.Any], bool]) -> 'PermissionInterface':
    """
    Decorator to create a PermissionInterface from a method/function.

    Example:
        class Permissions(ezperm.PermissionEnum):
            PERM1 = 'PERM1', 'Perm 1'
            PERM2 = 'PERM2', 'Perm 2'
            ...

            # ↓ HERE ↓ #
            @ezperm.permission
            def is_mage(obj: 'Hero') -> bool:
                return obj.type == 'mage'

    """
    class PermWrap(PermissionInterface):
        def _has_perm(self, obj: typing.Any) -> bool:
            return func(obj)
    return typing.cast(PermissionInterface, staticmethod(PermWrap()))



import dataclasses
import typing


class PermissionInterface:
    def _has_perm(self, obj: typing.Any, /) -> bool:
        """Check if the object has this permission. To be overriden in subclasses."""
        raise NotImplementedError

    def __call__(self, obj: typing.Any) -> bool:
        """Check if the object has this permission."""
        return self._has_perm(obj)

    def __or__(self, other: 'PermissionInterface') -> 'PermissionInterface':
        """
        Return a new permission that is a combination of this and other.
        This allows for the use of the | operator -> if one of the permissions is true, the whole expression is true.
        Example:
            permission_required = Perm1 | Perm2 | Perm3
        """
        @dataclasses.dataclass
        class OrPerm(PermissionInterface):
            perm1: 'PermissionInterface'
            perm2: 'PermissionInterface'

            def _has_perm(self, obj: typing.Any) -> bool:
                return self.perm1(obj) or self.perm2(obj)

        return OrPerm(self, other)

    def __and__(self, other: 'PermissionInterface') -> 'PermissionInterface':
        """
        Return a new permission that is a combination of this and other.
        This allows for the use of the & operator -> if both of the permissions are true, the whole expression is true.
        Example:
            permission_required = Perm1 & Perm2 & Perm3
        """
        @dataclasses.dataclass
        class AndPerm(PermissionInterface):
            perm1: 'PermissionInterface'
            perm2: 'PermissionInterface'

            def _has_perm(self, obj: typing.Any) -> bool:
                return self.perm1(obj) and self.perm2(obj)

        return AndPerm(self, other)

    def __invert__(self) -> 'PermissionInterface':
        """
        Return a new permission that is a negation of this permission.
        This allows for the use of the ~ operator -> if the permission is true, the whole expression is false.
        Example:
            permission_required = ~Perm1
        """
        @dataclasses.dataclass
        class NotPerm(PermissionInterface):
            perm: 'PermissionInterface'

            def _has_perm(self, obj: typing.Any) -> bool:
                return not self.perm(obj)

        return NotPerm(self)



import enum
import typing

from . import interfaces


class PermissionEnum(interfaces.PermissionInterface, enum.Enum):
    def _has_perm(self, obj: typing.Any) -> bool:
        """Check if the object has this permission."""
        raise NotImplementedError



try:
    import django
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Attempted import of `django` failed. Please specify `django` as an extra dependency of ezperm package.",
    )

from .models import *
